/**
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * Contact Information:<br>
 * BitCtrl Systems GmbH<br>
 * Wei&szlig;enfelser Stra&szlig;e 67<br>
 * 04229 Leipzig<br>
 * Phone: +49 341-490670<br>
 * mailto: info@bitctrl.de
 */

package org.geotools.data.vrz3;

import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;
import java.util.NoSuchElementException;

import org.geotools.data.FeatureReader;
import org.geotools.feature.AttributeType;
import org.geotools.feature.Feature;
import org.geotools.feature.FeatureType;
import org.geotools.feature.FeatureTypeBuilder;
import org.geotools.feature.SchemaException;
import org.geotools.feature.type.GeometricAttributeType;
import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryCollection;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.LinearRing;
import com.vividsolutions.jts.geom.MultiLineString;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.Polygon;

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.ReferenceValue;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.kex.wfs.wfs.ExportObjektDefinition;
import de.bsvrz.sys.funclib.bitctrl.geolib.WGS84Koordinate;
import de.bsvrz.sys.funclib.bitctrl.geolib.WGS84Punkt;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Klasse repr&auml;sentiert ein Feature des VRZ3DataStore.
 * 
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: VRZ3Feature.java 8125 2008-04-11 16:36:28Z gieseler $
 * 
 */
public class VRZ3Feature {

	/**
	 * FeatureReader Implementation
	 * 
	 * @author BitCtrl Systems GmbH, Gieseler
	 * @version $Id: VRZ3Feature.java 8125 2008-04-11 16:36:28Z gieseler $
	 * 
	 */
	private class Reader implements FeatureReader {

		/**
		 * das aktuelle Feature
		 */
		private Feature inputFeature = null;

		/**
		 * Puffer f&uuml;r die Attribute
		 */
		private Object[] inputBuffer = null;
		// private VRZ3ValueSetter[] fieldValueSetters;

		/**
		 * Liste aller Elemente des Typs
		 */
		List<SystemObject> objektliste = null;

		/**
		 * Iterator der Element-Liste
		 */
		ListIterator<SystemObject> li = null;

		/**
		 * @throws IOException
		 */
		public Reader() throws IOException {
			inputBuffer = new Object[_numAttribs];
			try {
				objektliste = _dav.getDataModel().getType(_davtypename)
						.getElements();
				li = objektliste.listIterator();
			} catch (Exception e) {
				throw new IOException(e.getMessage());
			}

			inputFeature = readFeature();
			DEBUG.info("Reader: " + _davtypename);
		}

		/**
		 * @param so
		 * @return
		 */
		private Geometry bestimmeGeometrieAusFlaechenKoordinaten(SystemObject so) {
			AttributeGroup ratg = _dav.getDataModel().getAttributeGroup(
					"atg.flchenKoordinaten"); //$NON-NLS-1$
			Data rd = so.getConfigurationData(ratg);

			if (rd != null) {
				List<WGS84Koordinate> koordinaten = getFlaechenKoordinaten(so);

				// Workaround fr Dummy-ISS, die nur eine Koordinate besitzen
				// createLineString erzeugt in diesem Fall einen Fehler
				if (koordinaten.size() == 1) {
					koordinaten.add(koordinaten.get(0));
				}

				Coordinate[] coords = new Coordinate[koordinaten.size()];

				for (int ii = 0; ii < koordinaten.size(); ii++) {
					coords[ii] = new Coordinate(
							koordinaten.get(ii).getLaenge(), koordinaten
									.get(ii).getBreite());
				}

				// geoserver mag die LinearRing's nicht
				// return _geomFactory.createLinearRing(coords);

				LinearRing shell = _geomFactory.createLinearRing(coords);
				return _geomFactory.createPolygon(shell, null);
			}

			// TODO Auto-generated method stub
			return null;
		}

		/**
		 * @param so
		 * @return
		 */
		private Geometry bestimmeGeometrieAusKomplexKoordinaten(SystemObject so) {
			AttributeGroup latg = _dav.getDataModel().getAttributeGroup(
					"atg.komplexKoordinaten"); //$NON-NLS-1$
			Data lrd = so.getConfigurationData(latg);
			List<Geometry> geometrien = new ArrayList<Geometry>();
			SystemObject refso;
			Data.ReferenceArray dra;

			if (lrd != null) {
				dra = lrd.getReferenceArray("PunktReferenz");
				for (int i = 0; i < dra.getLength(); i++) {
					refso = _dav.getDataModel().getObject(
							dra.getReferenceValue(i).getValueText());
					geometrien.add(bestimmeGeometrieAusPunktKoordinaten(refso));
				}

				dra = lrd.getReferenceArray("LinienReferenz");
				for (int i = 0; i < dra.getLength(); i++) {
					refso = _dav.getDataModel().getObject(
							dra.getReferenceValue(i).getValueText());
					geometrien
							.add(bestimmeGeometrieAusLinienKoordinaten(refso));
				}

				dra = lrd.getReferenceArray("FlchenReferenz");
				for (int i = 0; i < dra.getLength(); i++) {
					refso = _dav.getDataModel().getObject(
							dra.getReferenceValue(i).getValueText());
					geometrien
							.add(bestimmeGeometrieAusFlaechenKoordinaten(refso));
				}

			}

			return _geomFactory.createGeometryCollection(geometrien
					.toArray(new Geometry[geometrien.size()]));
		}

		/**
		 * @param so
		 * @return
		 */
		private Geometry bestimmeGeometrieAusLinienKoordinaten(SystemObject so) {
			AttributeGroup ratg = _dav.getDataModel().getAttributeGroup(
					"atg.linienKoordinaten"); //$NON-NLS-1$
			Data rd = so.getConfigurationData(ratg);
			if (rd != null) {
				List<WGS84Koordinate> koordinaten = getLinienKoordinaten(so);

				// Workaround fr Dummy-ISS, die nur eine Koordinate besitzen
				// createLineString erzeugt in diesem Fall einen Fehler
				if (koordinaten.size() == 1) {
					koordinaten.add(koordinaten.get(0));
				}

				Coordinate[] coords = new Coordinate[koordinaten.size()];

				for (int ii = 0; ii < koordinaten.size(); ii++) {
					coords[ii] = new Coordinate(
							koordinaten.get(ii).getLaenge(), koordinaten
									.get(ii).getBreite());
				}

				return _geomFactory.createLineString(coords);
			}

			return null;
		}

		/**
		 * @param so
		 * @return
		 */
		private Geometry bestimmeGeometrieAusLinienObjekten(SystemObject so) {
			Geometry geom = null;

			AttributeGroup latg = _dav.getDataModel().getAttributeGroup(
					"atg.bestehtAusLinienObjekten"); //$NON-NLS-1$
			Data lrd = so.getConfigurationData(latg);

			if (lrd != null) {
				Data.ReferenceArray lr = lrd
						.getReferenceArray("LinienReferenz");
				LineString[] lineStrings = new LineString[lr.getLength()];
				for (int i = 0; i < lr.getLength(); i++) {
					ReferenceValue rv = lr.getReferenceValue(i);
					String lrpid = rv.getValueText();
					// lineStrings[i] = _geomFactory.createLineString(coords);
					lineStrings[i] = (LineString) bestimmeGeometrieAusLinienKoordinaten(_dav
							.getDataModel().getObject(lrpid));
				}

				geom = _geomFactory.createMultiLineString(lineStrings);
			}

			return geom;
		}

		/**
		 * @param so
		 * @return
		 */
		private Geometry bestimmeGeometrieAusPunktKoordinaten(SystemObject so) {
			AttributeGroup ratg = _dav.getDataModel().getAttributeGroup(
					"atg.punktKoordinaten"); //$NON-NLS-1$
			Data rd = so.getConfigurationData(ratg);
			if (rd != null) {
				Coordinate coord = new Coordinate();
				coord.x = rd.getScaledValue("x").doubleValue();
				coord.y = rd.getScaledValue("y").doubleValue();

				return _geomFactory.createPoint(coord);
			}

			return null;
		}

		/**
		 * Bestimmt die Geometry zu einem Systemobjekt.
		 * 
		 * @param so
		 *            Systemobjekt
		 * @return Geometry Object
		 */

		private Geometry bestimmeGeometry(SystemObject so) {
			Geometry geom = null;

			// SystemObjectType stot =
			// _dav.getDataModel().getType("typ.linieXY");

			if (testeTyp(so.getType(), "typ.punktXY")) {
				geom = bestimmeGeometrieAusPunktKoordinaten(so);
			} else if (testeTyp(so.getType(), "typ.linieXY")) {
				geom = bestimmeGeometrieAusLinienKoordinaten(so);
			} else if (testeTyp(so.getType(), "typ.komplexXY")) {
				geom = bestimmeGeometrieAusKomplexKoordinaten(so);
			} else {
				// suche Geometrie-Attributgruppe
				for (VRZ3FeatureAttributDefinition fad : _featureDefinition) {
					if (fad.is_istGeometrie()) {
						if (fad._davatg.equals("atg.bestehtAusLinienObjekten")) {
							geom = bestimmeGeometrieAusLinienObjekten(so);
						} else if (fad._davatg.equals("atg.punktKoordinaten")) {
							geom = bestimmeGeometrieAusPunktKoordinaten(so);
						} else if (fad._davatg.equals("atg.flchenKoordinaten")) {
							geom = bestimmeGeometrieAusFlaechenKoordinaten(so);
						} else if (fad._davatg.equals("atg.komplexKoordinaten")) {
							geom = bestimmeGeometrieAusKomplexKoordinaten(so);
						}
					}
				}

				if (geom == null) {
					AttributeGroup latg = _dav.getDataModel()
							.getAttributeGroup("atg.bestehtAusLinienObjekten"); //$NON-NLS-1$
					Data lrd = so.getConfigurationData(latg);
					if (lrd != null) {
						geom = bestimmeGeometrieAusLinienObjekten(so);
					}
				}
			}

			return geom;
		}

		public void close() throws IOException {
			// TODO Auto-generated method stub

		}

		@Override
		protected void finalize() throws Throwable {
			super.finalize();
		}

		public FeatureType getFeatureType() {
			return _featureType;
		}

		/**
		 * @param so
		 * @return
		 */
		private List<WGS84Koordinate> getFlaechenKoordinaten(SystemObject so) {
			AttributeGroup ratg = _dav.getDataModel().getAttributeGroup(
					"atg.flchenKoordinaten"); //$NON-NLS-1$
			return getKoordinaten(so.getConfigurationData(ratg));
		}

		/**
		 * @param rd
		 * @return
		 */
		private List<WGS84Koordinate> getKoordinaten(Data rd) {
			List<WGS84Koordinate> kl = new ArrayList<WGS84Koordinate>();

			if (rd != null) {
				Data.NumberArray x = rd.getScaledArray("x"); //$NON-NLS-1$
				Data.NumberArray y = rd.getScaledArray("y"); //$NON-NLS-1$
				int anzahl = x.getLength();

				for (int i = 0; i < anzahl; i++) {
					kl.add(new WGS84Punkt(x.getValue(i).doubleValue(), y
							.getValue(i).doubleValue()));
				}
			}

			return kl;
		}

		// private Geometry
		// bestimmeGeometrieAusLinienObjekten_LineString(SystemObject so) {
		// Geometry geom = null;
		//	
		// AttributeGroup latg =
		// _dav.getDataModel().getAttributeGroup("atg.bestehtAusLinienObjekten");
		// //$NON-NLS-1$
		// Data lrd = so.getConfigurationData(latg);
		// List<WGS84Koordinate> koordinaten = new ArrayList<WGS84Koordinate>();
		//		
		// if( lrd != null ) {
		// Data.ReferenceArray lr = lrd.getReferenceArray("LinienReferenz");
		// LineString[] lineStrings = new LineString[lr.getLength()];
		// for( int i=0; i<lr.getLength(); i++) {
		// ReferenceValue rv = lr.getReferenceValue(i);
		// String lrpid = rv.getValueText();
		// koordinaten.addAll(getLinienKoordinaten(_dav.getDataModel().getObject(lrpid)));
		// }
		//			
		// // workaround fr Dummy-ISS, die nur eine Koordinate besitzen
		// // createLineString erzeugt in diesem Fall einen Fehler
		// if(koordinaten.size() == 1) {
		// koordinaten.add(koordinaten.get(0));
		// }
		//			
		// Coordinate[] coords = new Coordinate[koordinaten.size()];
		// for(int i=0; i<koordinaten.size(); i++) {
		// coords[i] = new Coordinate(koordinaten.get(i).get_laenge(),
		// koordinaten.get(i).get_breite());
		// }
		//
		// geom = _geomFactory.createLineString(coords);
		// }
		//		
		// return geom;
		// }

		/**
		 * @param so
		 * @return
		 */
		private List<WGS84Koordinate> getLinienKoordinaten(SystemObject so) {
			AttributeGroup ratg = _dav.getDataModel().getAttributeGroup(
					"atg.linienKoordinaten"); //$NON-NLS-1$
			return getKoordinaten(so.getConfigurationData(ratg));
		}

		public boolean hasNext() {
			return (inputFeature != null);
		}

		// Reads the next feature and returns the last one
		public Feature next() throws NoSuchElementException {
			if (inputFeature == null) {
				throw new NoSuchElementException("Reached the end of list");
			}

			Feature temp = inputFeature;

			try {
				inputFeature = readFeature();
			} catch (Exception e) {
				throw new NoSuchElementException(
						"Error retrieving next feature: " + e.toString());
			}

			return temp;
		}

		/**
		 * Liest ein Systemobjekt als Feature
		 * 
		 * @return Feature oder null, wenn am Ende der Liste
		 * 
		 * @throws IOException
		 */
		private Feature readFeature() throws IOException {
			Feature feature = null;
			if (!li.hasNext()) {
				return null;
			}

			SystemObject so = li.next();
			Geometry geom = bestimmeGeometry(so);

			try {
				inputBuffer[0] = geom;

				int index = 1;
				for (VRZ3FeatureAttribute a : _typAttribute) {
					if (index > inputBuffer.length) {
						System.err.println("Attributanzahl inkonsistent");
						throw new IOException("Attributanzahl inkonsistent");
					}

					inputBuffer[index++] = a.getAttributWert(so);
				}

				feature = _featureType.create(inputBuffer, so.getPid());
			} catch (Exception e) {
				throw new IOException("Exception building feature: "
						+ e.getMessage());
			}

			return feature;
		}
	}

	/**
	 * Repr&auml;sentiert die Definition eines Feature-Attributes
	 * 
	 * @author BitCtrl Systems GmbH, Gieseler
	 * @version $Id: VRZ3Feature.java 8125 2008-04-11 16:36:28Z gieseler $
	 * 
	 */
	private class VRZ3FeatureAttributDefinition {
		/**
		 * DaV-Attributgruppe
		 */
		String _davatg = null;

		/**
		 * DaV-Aspekt
		 */
		private String _davasp = null;

		/**
		 * legt fest, ob es sich bei dem Attribut um eine Geometrie handelt
		 */
		private boolean _istGeometrie = false;

		/**
		 * Erzeugt eine Instanz auf der Basis einer Attributgruppe und eines
		 * Aspekts
		 * 
		 * @param atg
		 *            Attributgruppe
		 * @param asp
		 *            Aspekt
		 */
		public VRZ3FeatureAttributDefinition(String atg, String asp) {
			_davatg = atg;
			_davasp = asp;
		}

		/**
		 * @return the _davasp
		 */
		public String get_davasp() {
			return _davasp;
		}

		/**
		 * @return the _davatg
		 */
		public String get_davatg() {
			return _davatg;
		}

		/**
		 * @return the _istGeometrie
		 */
		public boolean is_istGeometrie() {
			return _istGeometrie;
		}

		/**
		 * @param b
		 */
		public void set_istGeometrie(boolean b) {
			_istGeometrie = b;
		}
	}

	private final Debug DEBUG = Debug.getLogger();

	// private int _geomFieldIndex = -1;

	/** der zugeh&ouml;rige FeatureType */
	FeatureType _featureType = null;

	/** Anzahl der Attribute */
	int _numAttribs = 0;

	/** Namespace des Features */
	private URI _namespace = null;

	/** GeometryFactory */
	GeometryFactory _geomFactory = null;

	/** Datenverteiler-Verbindung */
	ClientDavConnection _dav;

	/** der Name des zugeh&ouml;rigen DaV-Datentyps */
	String _davtypename = null;

	/** DaV Typobjekt */
	private SystemObjectType _davSystemObjectType = null;

	/** Liste der Attrbut-Definitionen des Features */
	List<VRZ3FeatureAttributDefinition> _featureDefinition = new ArrayList<VRZ3FeatureAttributDefinition>();

	/** Liste der Attrbute des Features */
	List<VRZ3FeatureAttribute> _typAttribute = new ArrayList<VRZ3FeatureAttribute>();

	/** DaV-Datenmodell */
	private DataModel _config = null;

	/** Name des Geometry-Attributes */
	private final String _geometryName = "geometrie";

	/** Name des Feature-Typs */
	private String _typeName;

	/**
	 * Konstruktor
	 * 
	 * @param dav
	 *            Datenverteiler-Verbindung
	 * @param objects
	 *            Feld mit Objektbeschreibungen
	 * @param namespace
	 *            Namespace
	 */
	public VRZ3Feature(ClientDavConnection dav, Object[] objects, URI namespace) {
		_namespace = namespace;
		_dav = dav;
		_config = _dav.getDataModel();
		_davtypename = (String) objects[ExportObjektDefinition.TYP];
		_featureDefinition.add(new VRZ3FeatureAttributDefinition(
				(String) objects[ExportObjektDefinition.ATG],
				(String) objects[ExportObjektDefinition.ASP]));
		_geomFactory = new GeometryFactory();
	}

	/**
	 * Bestimmt die Attribute des Features.
	 */
	private void bestimmeAttribute() {
		ListIterator<VRZ3FeatureAttributDefinition> li = _featureDefinition
				.listIterator();
		// List<AttributeType> atts = new ArrayList<AttributeType>();

		// erstes Attribut ist der Name
		_typAttribute.addAll(VRZ3AttributeFactory.erzeugeObjektAttribute());

		while (li.hasNext()) {
			VRZ3FeatureAttributDefinition fa = li.next();

			if (!fa.is_istGeometrie()) {
				_typAttribute.addAll(VRZ3AttributeFactory.erzeugeAttribute(
						_dav, fa._davatg));
				// atts.addAll(fa.bestimmeAttribute(_dav));
			}
		}
	}

	/**
	 * Bestimmt den Geometrie-Typ des Features: Wenn der Typ von einem der Typen
	 * des TeilmodellGeoReferenzierungGlobal abgeleitet ist, wird der
	 * Geoemtrietyp aus dem entsprechenden Basistyp gebildet. Konnte damit keine
	 * Geometrie gefunden werden, werden die f&uuml;r den Typ zu exportierenden
	 * Attributgruppen nach folgenden Attributgruppen durchsucht:
	 * atg.bestehtAusLinienObjekten atg.punktKoordinaten
	 * 
	 * Wird eine dieser Attributgruppen
	 * 
	 * Die zur Bildung der Geometrie zu verwendete Attributgruppe wird markiert,
	 * um sie von der Bildung der weiteren Attribute des Featuretyps
	 * auszuschlie&szlig;en.
	 * 
	 * @return AttributeType
	 * @throws SchemaException
	 */
	private AttributeType bestimmeGeometrieTyp() throws SchemaException {
		Class<?> geomType = null;
		String geometrieatg = null;

		if (testeTyp(_davSystemObjectType, "typ.punktXY")) {
			geomType = Point.class;
			geometrieatg = "atg.punktKoordinaten";
		}

		if (testeTyp(_davSystemObjectType, "typ.linieXY")) {
			geomType = LineString.class;
			geometrieatg = "atg.linienKoordinaten";
		}

		if (testeTyp(_davSystemObjectType, "typ.flcheXY")) {
			geomType = Polygon.class;
			geometrieatg = "atg.flchenKoordinaten";
		}

		if (testeTyp(_davSystemObjectType, "typ.komplexXY")) {
			geomType = GeometryCollection.class;
			geometrieatg = "atg.komplexKoordinaten";
		}

		if (geomType != null) {
			setzeGeometrieAttribut(geometrieatg);
			return erzeugeGeometrieAttributTyp(geomType);
		}

		// ueberprfe, ob eine der konfigurierten Attributgruppen eine Geometrie
		// reprsentiert
		for (VRZ3FeatureAttributDefinition d : _featureDefinition) {
			if (d.get_davatg().equals("atg.bestehtAusLinienObjekten")) {
				d.set_istGeometrie(true);
				geomType = MultiLineString.class;
			}

			if (d.get_davatg().equals("atg.punktKoordinaten")) {
				d.set_istGeometrie(true);
				geomType = Point.class;
			}
		}

		if (geomType == null) {
			throw new SchemaException(
					"Der Geometrietyp kann nicht bestimmt werden");
		}

		return erzeugeGeometrieAttributTyp(geomType);
	}

	/**
	 * Erzeugt das Feature.
	 * 
	 * @throws IOException
	 * @throws SchemaException
	 */
	public void buildFeature() throws SchemaException {
		DEBUG.info("Bilde Feature: " + _davtypename);
		_davSystemObjectType = _config.getType(_davtypename);

		// zuerst Geometrie suchen
		AttributeType geometrieatt = bestimmeGeometrieTyp();

		// die weiteren Attribute bestimmen
		bestimmeAttribute();

		try {
			String typeName = _davSystemObjectType.getName();
			if (typeName.indexOf(".") >= 0) {
				typeName = typeName.substring(0, typeName.indexOf("."));
			}

			_typeName = umlauteUmwandeln(typeName);

			FeatureTypeBuilder builder = FeatureTypeBuilder
					.newInstance(_typeName);

			builder.setNamespace(_namespace);

			if (geometrieatt != null) {
				builder.addType(geometrieatt);
			}

			for (VRZ3FeatureAttribute att : _typAttribute) {
				builder.addType(att.get_geotoolsAttributeType());
			}

			setSchema(builder.getFeatureType());
			DEBUG.info("Bilde Feature beendet: " + _typeName);
		} catch (SchemaException schexp) {
			throw new SchemaException("Exception creating feature type: "
					+ schexp.toString());
		}

	}

	/**
	 * Erzeugt das Geometrie-Attribut.
	 * 
	 * @param geomType
	 *            Geometrie-Typ
	 * @return Attribut
	 */
	private AttributeType erzeugeGeometrieAttributTyp(Class<?> geomType) {
		CoordinateReferenceSystem crs = DefaultGeographicCRS.WGS84;

		GeometricAttributeType refGeom = new GeometricAttributeType(
				_geometryName, geomType, true, null, crs, null);

		// return AttributeTypeFactory.newAttributeType(_geometryName ,
		// geomType, true);

		return refGeom;
	}

	/**
	 * Erzeugt einen FeatureReader.
	 * 
	 * @return FeatureReader
	 */
	public FeatureReader getFeatureReader() {
		try {
			return new Reader();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Gibt das Schema zur&uuml;ck
	 * 
	 * @return den mit dem Feature assoziierten FeatureType
	 */
	public FeatureType getSchema() {
		DEBUG.info("getSchema: " + _davtypename);
		return _featureType;
	}

	// private List<AttributeType> getAttributes(Data data) {
	// List<AttributeType> atts = new ArrayList<AttributeType>();
	//		
	// // Geometrie
	// CoordinateReferenceSystem crs = DefaultGeographicCRS.WGS84;
	//        
	// GeometricAttributeType refGeom = new GeometricAttributeType(
	// "the_geom", com.vividsolutions.jts.geom.Geometry.class, true, null, crs,
	// null);
	//    	
	// atts.add(refGeom);
	//		
	// if (data.isPlain()) {
	// atts.add(getAttribute(data));
	// } else if (data.isArray()) {
	// /* Iterator zum Durchlaufen des Feldes */
	// Iterator i = data.iterator();
	//
	// /* Name dieses Datums */
	// // String name = data.getName();
	//
	// while (i.hasNext()) {
	// atts.add(getAttribute((Data) i.next()));
	// }
	// } else if (data.isList()) {
	// /* Iterator zum Durchlaufen der Liste */
	// Iterator i = data.iterator();
	//
	// /* Name dieses Datums */
	// // String name = data.getName();
	//
	// while (i.hasNext()) {
	// atts.add(getAttribute((Data) i.next()));
	// }
	// }
	//		
	// return atts;
	// }

	// private List<AttributeType> getAttributeValues(Data data) {
	// List<AttributeType> atts = new ArrayList<AttributeType>();
	//		
	// // Geometrie
	// CoordinateReferenceSystem crs = DefaultGeographicCRS.WGS84;
	//        
	// GeometricAttributeType refGeom = new GeometricAttributeType(
	// "the_geom", com.vividsolutions.jts.geom.Geometry.class, true, null, crs,
	// null);
	//    	
	// atts.add(refGeom);
	//		
	// if (data.isPlain()) {
	// atts.add(getAttribute(data));
	// } else if (data.isArray()) {
	// /* Iterator zum Durchlaufen des Feldes */
	// Iterator i = data.iterator();
	//
	// /* Name dieses Datums */
	// // String name = data.getName();
	//
	// while (i.hasNext()) {
	// atts.add(getAttribute((Data) i.next()));
	// }
	// } else if (data.isList()) {
	// /* Iterator zum Durchlaufen der Liste */
	// Iterator i = data.iterator();
	//
	// /* Name dieses Datums */
	// // String name = data.getName();
	//
	// while (i.hasNext()) {
	// atts.add(getAttribute((Data) i.next()));
	// }
	// }
	//		
	// return atts;
	// }
	//
	//	

	/**
	 * F&uuml;gt eine Attributgruppe zum Feature hinzu.
	 * 
	 * @param objects
	 *            Feld mit Objektbeschreibungen
	 */
	public void hinzufuegenAttributgruppe(final Object[] objects) {
		_featureDefinition.add(new VRZ3FeatureAttributDefinition(
				(String) objects[ExportObjektDefinition.ATG],
				(String) objects[ExportObjektDefinition.ASP]));
	}

	/**
	 * Setzt das Schema
	 * 
	 * @param ft
	 *            Feature-Typ
	 */
	private void setSchema(FeatureType ft) {
		_featureType = ft;

		_numAttribs = _featureType.getAttributeCount();
		// _geomFieldIndex = -1;
	}

	// private GeometricAttributeType bestimmeGeometrie(VRZ3FeatureAttribute
	// att) {
	// Geometry geom = null;
	// // Geometrie
	// CoordinateReferenceSystem crs = DefaultGeographicCRS.WGS84;
	//       	
	// String geometrieatg = null;
	//        
	// GeometricAttributeType refGeom = new GeometricAttributeType(
	// "the_geom", com.vividsolutions.jts.geom.Geometry.class, true, null, crs,
	// null);
	//    	
	// //SystemObjectType sott = _dav.getDataModel().getType("typ.linieXY");
	//        
	// if(
	// _davSystemObjectType.getType().inheritsFrom(_config.getType("typ.linieXY"))
	// ) {
	// geometrieatg = "atg.linienKoordinaten";
	// AttributeGroup atg = _dav.getDataModel().getAttributeGroup(geometrieatg);
	// //$NON-NLS-1$
	// Data data = _davSystemObjectType.getConfigurationData(atg);
	// if( data != null ) {
	// Data.NumberArray x = data.getScaledArray("x"); //$NON-NLS-1$
	// Data.NumberArray y = data.getScaledArray("y"); //$NON-NLS-1$
	// int anzahl = x.getLength();
	//
	// Coordinate[] coords = new Coordinate[anzahl];
	// for(int i=0; i<anzahl; i++) {
	// coords[i] = new Coordinate(x.getValue(i).doubleValue(),
	// y.getValue(i).doubleValue());
	// }
	//				
	// geom = _geomFactory.createLineString(coords);
	// }
	// }
	// else {
	// AttributeGroup atg =
	// _dav.getDataModel().getAttributeGroup("atg.bestehtAusLinienObjekten");
	// //$NON-NLS-1$
	// Data lrd = _davSystemObjectType.getConfigurationData(atg);
	// if( lrd != null ) {
	// geom = bestimmeGeometrieAusLinienObjekten(_davSystemObjectType);
	// }
	// }
	//        
	// }

	/**
	 * @param geometrieatg
	 */
	private void setzeGeometrieAttribut(String geometrieatg) {
		for (VRZ3FeatureAttributDefinition d : _featureDefinition) {
			if (d.get_davatg().equals(geometrieatg)) {
				d.set_istGeometrie(true);
			}
		}
	}

	// AttributeType getAttribute( Data data ) {
	// Class<?> typeClass;
	// String name = data.getName();
	// String defaultvalue = null;
	// de.bsvrz.dav.daf.main.config.AttributeType type =
	// data.getAttributeType();
	//	
	// if (type instanceof StringAttributeType) {
	// typeClass = String.class;
	// } else if (type instanceof IntegerAttributeType) {
	// typeClass = Integer.class;
	// IntegerValueRange wertebereich = ((IntegerAttributeType)
	// type).getRange();
	// defaultvalue = ((IntegerAttributeType) type).getDefaultAttributeValue();
	// } else if (type instanceof DoubleAttributeType) {
	// typeClass = Double.class;
	// } else if (type instanceof TimeAttributeType) {
	// typeClass = Integer.class;
	// } else if (type instanceof ReferenceAttributeType) {
	// typeClass = String.class;
	// // result = new ObjektReferenz(data.getName());
	// // SystemObject ref = data.asReferenceValue().getSystemObject();
	// // if (ref == null) {
	// // ((ObjektReferenz) result).set(0L);
	// // } else {
	// // ((ObjektReferenz) result).set(ref.getId());
	// // }
	// }
	// else
	// return null;
	//		
	// // AttributeType at = AttributeTypeFactory.newAttributeType(name,
	// // typeClass, false, size, defaultvalue);
	//		
	// AttributeType at = AttributeTypeFactory.newAttributeType(name,
	// typeClass, false, 0, defaultvalue);
	//		
	// return at;
	// }

	// private void fetchData(Data data, List<String> referenceValues, int
	// level,
	// StringBuffer sb) {
	// if (data.isPlain()) {
	// /* Name dieses Datums */
	// String name = data.getName();
	// de.bsvrz.dav.daf.main.config.AttributeType attributeType =
	// data.getAttributeType();
	// String value;
	//
	// if (attributeType instanceof ReferenceAttributeType) {
	// value = referenceValues.get(0);
	// referenceValues.remove(0);
	// } else {
	// /* Der konkrete Wert dieses Datums */
	// value = data.asTextValue().getValueText();
	// }
	// // <datum name="name" wert="value"/>
	// sb.append(indent(level));
	// sb.append("<datum name=\"");
	// sb.append(name);
	// sb.append("\"" + " wert=\"");
	// sb.append(value);
	// sb.append("\"/>\n");
	// } else if (data.isArray()) {
	// /* Iterator zum Durchlaufen des Feldes */
	// Iterator i = data.iterator();
	//
	// /* Name dieses Datums */
	// String name = data.getName();
	//
	// // <datenfeld name="name">
	// sb.append(indent(level)).append("<datenfeld name=\"").append(name)
	// .append("\">\n");
	//
	// level++;
	// while (i.hasNext()) {
	// printData((Data) i.next(), referenceValues, level, sb);
	// }
	// level--;
	// // </datenfeld>
	// sb.append(indent(level)).append("</datenfeld>\n");
	// } else if (data.isList()) {
	// /* Iterator zum Durchlaufen der Liste */
	// Iterator i = data.iterator();
	//
	// /* Name dieses Datums */
	// String name = data.getName();
	//
	// if (level == 3) {
	// // ffnendes Tag fr Datensatz schreiben
	// sb.append(indent(level)).append("<datensatz pid=\"").append(
	// name).append("\">\n");
	// } else {
	// // <datenliste name="name">
	// sb.append(indent(level)).append("<datenliste name=\"").append(
	// name).append("\">\n");
	// }
	//
	// level++;
	// while (i.hasNext()) {
	// printData((Data) i.next(), referenceValues, level, sb);
	// }
	// level--;
	//
	// if (level == 3) {
	// // Schlieendes Tag fr Datensatz schreiben
	// sb.append(indent(level)).append("</datensatz>\n");
	// } else {
	// // </datenliste>
	// sb.append(indent(level)).append("</datenliste>\n");
	// }
	// }
	// }

	/**
	 * Test auf abgeleiteten Typ.
	 * 
	 * @param dt
	 * @param totest
	 * @return treu, wenn der Typ gleich oder abgeleitet ist, sonst false.
	 */
	boolean testeTyp(SystemObjectType dt, String totest) {
		SystemObjectType stot = _dav.getDataModel().getType(totest);

		return dt.equals(stot) || dt.inheritsFrom(stot);
	}

	// /**
	// * Gets the ValueSetters
	// *
	// * @return An array of valueSetters to be used for IO operations
	// *
	// * @throws SchemaException An attribute of an unsupported type was found.
	// */
	// private VRZ3ValueSetter[] getValueSetters() throws SchemaException {
	// VRZ3ValueSetter[] fieldValueSetters = new VRZ3ValueSetter[_numAttribs];
	//
	// for (int i = 0; i < _featureType.getAttributeCount(); i++) {
	// AttributeType at = _featureType.getAttributeType(i);
	// Class atc = at.getType();
	//
	// if (i == _geomFieldIndex) {
	// fieldValueSetters[i] = null;
	// } else if (atc == Integer.class) {
	// fieldValueSetters[i] = new VRZ3ValueSetter("0") {
	// @Override
	// protected void stringToValue()
	// throws Exception {
	// objValue = new Integer(strValue);
	// }
	// };
	// } else if (atc == Double.class) {
	// fieldValueSetters[i] = new VRZ3ValueSetter("0") {
	// @Override
	// protected void stringToValue()
	// throws Exception {
	// objValue = new Double(strValue);
	// }
	//
	// protected void valueToString() {
	// // TODO use DecimalFormat class!!!
	// super.valueToString();
	// }
	// };
	// } else if (atc == Float.class) {
	// fieldValueSetters[i] = new VRZ3ValueSetter("0") {
	// @Override
	// protected void stringToValue()
	// throws Exception {
	// objValue = new Float(strValue);
	// }
	//
	// @Override
	// protected void valueToString() {
	// // TODO use DecimalFormat class!!!
	// super.valueToString();
	// }
	// };
	// } else if (atc == Boolean.class) {
	// fieldValueSetters[i] = new VRZ3ValueSetter("false") {
	// @Override
	// protected void stringToValue()
	// throws Exception {
	// objValue = new Boolean("T".equalsIgnoreCase(
	// strValue) ? "true"
	// : ("F".equalsIgnoreCase(
	// strValue) ? "false" : strValue));
	// }
	//
	// @Override
	// protected void valueToString() {
	// if ((objValue == null)
	// || (((Boolean) objValue).booleanValue() == false)) {
	// strValue = "F";
	// } else {
	// strValue = "T";
	// }
	// }
	// };
	// } else if (atc == Date.class) {
	// // TODO Check conversion of date values - switch to java.sql.Date
	// fieldValueSetters[i] = new VRZ3ValueSetter("") {
	// protected SimpleDateFormat dateFormat = new SimpleDateFormat(
	// "yyyyMMdd");
	//
	// @Override
	// protected void stringToValue()
	// throws Exception {
	// if ((strValue != null) && !strValue.equals("")) {
	// objValue = dateFormat.parse(strValue);
	// } else {
	// objValue = null;
	// }
	//
	// // Date.valueOf(strValue.substring(0, 4) + "-" + strValue.substring(4, 6)
	// + "-" + strValue.substring(6));
	// }
	//
	// @Override
	// protected void valueToString() {
	// if (objValue == null) {
	// strValue = "";
	// } else {
	// strValue = dateFormat.format(objValue);
	//
	// // strValue = ((Date) objValue).getYear() + "" + ((Date)
	// objValue).getMonth() + "" + ((Date) objValue).getDay();
	// }
	// }
	// };
	// } else if (atc == String.class) {
	// fieldValueSetters[i] = new VRZ3ValueSetter("") {
	// @Override
	// protected void stringToValue()
	// throws Exception {
	// objValue = new String(strValue);
	// }
	//
	// // Quotes the string
	// @Override
	// protected void valueToString() {
	// strValue = new String("\""
	// + objValue.toString().replaceAll("\"",
	// "\"\"") + "\"");
	// }
	// };
	// } else {
	// throw new SchemaException("Unsupported attribute type: "
	// + atc.getName());
	// }
	// }
	//
	// return fieldValueSetters;
	// }

	/**
	 * Wandelt den Namen des Features in einen Namen ohne Umlaute um
	 * 
	 * @param typename
	 *            Feature-Name
	 * @return umgewandelter Name
	 */
	private String umlauteUmwandeln(final String typename) {
		String newtypename = typename;

		newtypename = newtypename.replaceAll("", "ae");
		newtypename = newtypename.replaceAll("", "Ae");
		newtypename = newtypename.replaceAll("", "oe");
		newtypename = newtypename.replaceAll("", "Ue");
		newtypename = newtypename.replaceAll("", "ue");
		newtypename = newtypename.replaceAll("", "Ue");
		newtypename = newtypename.replaceAll("", "ss");

		return (newtypename);
	}
}
