/**
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * Contact Information:<br>
 * BitCtrl Systems GmbH<br>
 * Wei&szlig;enfelser Stra&szlig;e 67<br>
 * 04229 Leipzig<br>
 * Phone: +49 341-490670<br>
 * mailto: info@bitctrl.de
 */

package org.geotools.data.vrz3;

/**
 * Feature-Attribut der VRZ3-Features
 * 
 * @author BitCtrl Systems GmbH, Gieseler
 * @version $Id: VRZ3ValueSetter.java 6072 2008-01-30 17:11:25Z gieseler $
 *
 */

public abstract class VRZ3ValueSetter {
	/** The object value as string	*/
    protected String strValue = null; 
    
    /** the object value	*/
    protected Object objValue = null;
    
    /** String representation of the default value (must be correctly converted into object by the stringToValue method!!!) */
    private String defaultValue = "";
    
    /** Fehlermeldung */
    private String errorMessage = "";

    /**
     * <p>The constructor accepts a default value for the ValueSetter.</p>
     *
     * @param defa String representation of the default value
     */
    public VRZ3ValueSetter(String defa) {
        defaultValue = defa;
    }

    /**
     * <p>Sets the value as a String. After a setString call, getValue() can be
     * used to access the converted value.</p>
     *
     * @param value String representation of the object value
     *
     * @return true if conversion was successful
     */
    public final boolean setString(String value) {
        strValue = value;

        try {
            stringToValue();

            return true;
        } catch (Exception e) {
            errorMessage = e.getMessage();
            strValue = defaultValue;

            try {
                stringToValue();
            } catch (Exception ex) {
                // Should never reach this place!!!!
                objValue = null;
                errorMessage += ". Bad default string value";
            }

            return false;
        }
    }

    /**
     * <p>Returns the string value.</p>
     * @return String-Wert
     *
     */
    public final String getString() {
        return strValue;
    }

    /**
     * <p>Sets the object value, and calculates the String value.</p>
     *
     * @param value The Object value
     */
    public final void setValue(Object value) {
        objValue = value;
        valueToString();
    }

    /**
     * <p>Gets the object value.</p>
     * @return 
     *
     */
    public final Object getValue() {
        return objValue;
    }

    /**
     * <p>Gets and resets the current error message.</p>
     *
     * @return The current error message, "" if none
     */
    public final String getError() {
        String tmp = errorMessage;
        errorMessage = "";

        return tmp;
    }

    /**
     * <p>Converts an object value to string - the default implementation uses
     * toString for non-null values.</p>
     */
    protected void valueToString() {
        // String.valueOf() would yeld "null"
        if (objValue == null) {
            strValue = "";
        } else {
            strValue = objValue.toString();
        }
    }

    /**
     * <p>This method must be overridden by descendants in order to implement the
     * correct conversion between strings and object values. <br>
     * Must throw an exception if conversion failed</p>.
     *
     * @throws Exception if value conversion failed
     */
    protected abstract void stringToValue() throws Exception;
}
